﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Reflection;
using System.Runtime.Remoting.Lifetime;
using System.IO;

namespace DynamicLoad
{
    public class AssemblyLoader : IDisposable
    {
        private readonly object _lockThis = new object();
        private Dictionary<string, Queue<InvokerInfo>> _caches =
            new Dictionary<string, Queue<InvokerInfo>>();

        //加载所有可用的程序集
        /// <summary>
        /// 加载所有可用的程序集
        /// </summary>
        /// <param name="dlls"></param>
        public void LoadAssemblys(Dictionary<string, string> dlls)
        {
            foreach (KeyValuePair<string, string> kvp in dlls)
                LoadAssembly(kvp.Key, kvp.Value);
        }
        //卸载程序集
        /// <summary>
        /// 卸载程序集
        /// </summary>
        /// <param name="dllName"></param>
        public void Unload(string dllName)
        {
            InvokerInfo info = _caches[dllName].Dequeue();
            AppDomain.Unload(info.Domain);
        }
        //调用指定程序集中的指定方法
        /// <summary>
        /// 调用指定程序集中指定的方法
        /// </summary>
        /// <param name="dllName">程序集名称</param>
        /// <param name="methodName">方法名称</param>
        /// <param name="methodParams">参数数组</param>
        /// <returns></returns>
        public object InvokeMethod(string dllName, string methodName, object[] methodParams)
        {
            object result;
            InvokerInfo info;
            lock (_lockThis)
            {
                info = GetInvoker(dllName);
                info.Ref++;
            }

            result = info.Invoker.InvokeMethod(methodName, methodParams);

            lock (_lockThis)
            {
                info.Ref--;
                TryToUnLoad(dllName, info);
            }

            return result;
        }

        //加载指定的程序集
        /// <summary>
        /// 加载指定的程序集
        /// </summary>
        /// <param name="dllName"></param>
        /// <param name="typeName"></param>
        private void LoadAssembly(string dllName, string typeName)
        {
            //Get object from cache
            Queue<InvokerInfo> result;
            _caches.TryGetValue(dllName, out result);

            if (result == null || result.Count == 0)
            {
                //Get TimeStamp of file
                FileInfo info =
                    new FileInfo(
                        AppDomain.CurrentDomain.BaseDirectory + @"ServiceDlls\" + dllName);

                if (!info.Exists)
                    throw new Exception(AppDomain.CurrentDomain.BaseDirectory 
                        + @"ServiceDlls\" + dllName + " not exist");

                CacheMethodInvoker(dllName, typeName, info.LastWriteTime);
            }
        }
        //缓存指定的方法调用信息
        /// <summary>
        /// 缓存指定的方法调用信息
        /// </summary>
        /// <param name="dllName"></param>
        /// <param name="typeName"></param>
        /// <param name="lastWriteTime"></param>
        /// <returns></returns>
        private InvokerInfo CacheMethodInvoker(string dllName, string typeName, DateTime lastWriteTime)
        {
            MethodInvoker invoker;

            var invokerInfo = new InvokerInfo();

            var setup = new AppDomainSetup
            {
                ShadowCopyFiles = "true",
                ShadowCopyDirectories = AppDomain.CurrentDomain.BaseDirectory + @"ServiceDlls\",
                ConfigurationFile = "DynamicLoadAssembly.exe.config",
                ApplicationBase = AppDomain.CurrentDomain.BaseDirectory
            };

            AppDomain domain = AppDomain.CreateDomain(dllName, null, setup);

            domain.DoCallBack(delegate { LifetimeServices.LeaseTime = TimeSpan.Zero; });

            invokerInfo.Domain = domain;
            invokerInfo.LastWriteTime = lastWriteTime;
            invokerInfo.TypeName = typeName;

            BindingFlags bindings = 
                BindingFlags.CreateInstance | BindingFlags.Instance | BindingFlags.Public;
            object[] para = new object[] { setup.ShadowCopyDirectories + @"\" + dllName, typeName };
            try
            {
                invoker = (MethodInvoker)domain.CreateInstanceFromAndUnwrap(
                                             Assembly.GetExecutingAssembly().CodeBase.Substring(8), 
                                             typeof(MethodInvoker).FullName,
                                             true, bindings, null, para, null,
                                             null, null);
            }
            catch (Exception ex)
            {
                throw new Exception(
                    "Can't create object which type is " + typeof(MethodInvoker).FullName + " from assembly " +
                    Assembly.GetExecutingAssembly().CodeBase + ",Error Message: " + ex.Message);
            }

            if (invoker == null)
                throw new Exception(
                    "Can't find type " + typeof(MethodInvoker).FullName + " from " +
                    Assembly.GetExecutingAssembly().CodeBase);

            try
            {
                invoker.LoadAllMethods();
            }
            catch (Exception ex)
            {
                throw new Exception("Can't initialize object which type is " + typeof(MethodInvoker).FullName +
                                    " from " +
                                    Assembly.GetExecutingAssembly().CodeBase + ",Error Message: " + ex.Message);
            }

            invokerInfo.Invoker = invoker;
            invokerInfo.Ref = 0;

            if (_caches.Keys.Contains(dllName))
            {
                _caches[dllName].Enqueue(invokerInfo);
            }
            else
            {
                Queue<InvokerInfo> queue = new Queue<InvokerInfo>();
                queue.Enqueue(invokerInfo);
                _caches[dllName] = queue;
            }

            return invokerInfo;
        }
        //尝试卸载程序集
        /// <summary>
        /// 尝试卸载程序集
        /// </summary>
        /// <param name="dllName"></param>
        /// <param name="currentInfo"></param>
        private void TryToUnLoad(string dllName, InvokerInfo currentInfo)
        {
            InvokerInfo info = _caches[dllName].Peek();

            if (info == currentInfo)
                return;

            if (info.Ref == 0)
            {
                Unload(dllName);
            }
        }
        //获取指定程序集的调用信息
        /// <summary>
        /// 获取指定程序集的调用信息
        /// </summary>
        /// <param name="dllName"></param>
        /// <returns></returns>
        private InvokerInfo GetInvoker(string dllName)
        {
            //Get object from cache
            Queue<InvokerInfo> result;
            _caches.TryGetValue(dllName, out result);

            if (result == null)
            {
                throw new Exception(dllName + " not loaded");
            }

            //Get TimeStamp of file
            FileInfo info =
                new FileInfo(AppDomain.CurrentDomain.BaseDirectory + @"ServiceDlls\" + dllName);

            if (!info.Exists)
            {
                return result.ToArray()[result.Count - 1];
            }

            if (info.LastWriteTime > result.ToArray()[result.Count - 1].LastWriteTime)
            {
                return CacheMethodInvoker(dllName, result.Peek().TypeName, info.LastWriteTime);
            }

            return result.ToArray()[result.Count - 1];
        }

        #region IDisposable 成员

        public void Dispose()
        {
            _caches.Clear();

            foreach (var o in _caches.Keys)
            {
                Unload(o);
            }
        }

        #endregion
    }
}
